@@ -12,53 +12,66 @@ import (
 
 	"github.com/AdguardTeam/golibs/errors"
 	"github.com/AdguardTeam/golibs/log"
-	"golang.org/x/net/http2"
 )
 
-// NextProtoDQ is the ALPN token for DoQ. During connection establishment,
-// DNS/QUIC support is indicated by selecting the ALPN token "dq" in the
+// NextProtoDQ is the ALPN token for DoQ. During the connection establishment,
+// DNS/QUIC support is indicated by selecting the ALPN token "doq" in the
 // crypto handshake.
-// Current draft version:
-// https://datatracker.ietf.org/doc/html/draft-ietf-dprive-dnsoquic-02
-const NextProtoDQ = "doq-i02"
+// The current draft version is https://datatracker.ietf.org/doc/rfc9250/.
+const NextProtoDQ = "doq"
 
 // compatProtoDQ is a list of ALPN tokens used by a QUIC connection.
 // NextProtoDQ is the latest draft version supported by dnsproxy, but it also
 // includes previous drafts.
-var compatProtoDQ = []string{NextProtoDQ, "doq-i00", "dq", "doq"}
+var compatProtoDQ = []string{NextProtoDQ, "doq-i00", "dq", "doq-i02"}
 
-// RootCAs is the CertPool that must be used by all upstreams
-// Redefining RootCAs makes sense on iOS to overcome the 15MB memory limit of the NEPacketTunnelProvider
-// nolint
+// RootCAs is the CertPool that must be used by all upstreams. Redefining
+// RootCAs makes sense on iOS to overcome the 15MB memory limit of the
+// NEPacketTunnelProvider.
 var RootCAs *x509.CertPool
 
-// CipherSuites - custom list of TLSv1.2 ciphers
-// nolint
+// CipherSuites is a custom list of TLSv1.2 ciphers.
 var CipherSuites []uint16
 
-// TODO: refactor bootstrapper, it's overcomplicated and hard to understand what it does
+// TODO(ameshkov): refactor bootstrapper, it's overcomplicated and hard to
+// understand what it does.
 type bootstrapper struct {
-	URL            *url.URL
-	resolvers      []*Resolver // list of Resolvers to use to resolve hostname, if necessary
-	dialContext    dialHandler // specifies the dial function for creating unencrypted TCP connections.
-	resolvedConfig *tls.Config
-	sync.RWMutex
+	// URL is the upstream server address.
+	URL *url.URL
 
-	// stores options for AddressToUpstream func:
-	// callbacks for checking certificates, timeout,
-	// the need to verify the server certificate,
-	// the addresses of upstream servers, etc
+	// resolvers is a list of *net.Resolver to use to resolve the upstream
+	// hostname, if necessary.
+	resolvers []*Resolver
+
+	// dialContext is the dial function for creating unencrypted TCP
+	// connections.
+	dialContext dialHandler
+
+	// resolvedConfig is a *tls.Config that is used for encrypted DNS protocols.
+	resolvedConfig *tls.Config
+
+	// sessionsCache is necessary to achieve TLS session resumption.  We create
+	// once when the bootstrapper is created and re-use every time when we need
+	// to create a new tls.Config.
+	sessionsCache tls.ClientSessionCache
+
+	// guard protects dialContext and resolvedConfig.
+	guard sync.RWMutex
+
+	// options is the Options that were passed to the AddressToUpstream
+	// function.  It configures different upstream properties: callbacks for
+	// checking certificates, timeout, etc.
 	options *Options
 }
 
-// newBootstrapperResolved creates a new bootstrapper that already contains resolved config.
-// This can be done only in the case when we already know the resolver IP address.
-// options -- Upstream customization options
+// newBootstrapperResolved creates a new bootstrapper that already contains
+// resolved config. This can be done only in the case when we already know the
+// resolver IP address passed via options.
 func newBootstrapperResolved(upsURL *url.URL, options *Options) (*bootstrapper, error) {
 	// get a host without port
 	host, port, err := net.SplitHostPort(upsURL.Host)
 	if err != nil {
-		return nil, fmt.Errorf("bootstrapper requires port in address %s", upsURL.String())
+		return nil, fmt.Errorf("bootstrapper requires port in address %s", upsURL)
 	}
 
 	var resolverAddresses []string
@@ -70,6 +83,10 @@ func newBootstrapperResolved(upsURL *url.URL, options *Options) (*bootstrapper,
 	b := &bootstrapper{
 		URL:     upsURL,
 		options: options,
+		// Use the default capacity for the LRU cache.  It may be useful to
+		// store several caches since the user may be routed to different
+		// servers in case there's load balancing on the server-side.
+		sessionsCache: tls.NewLRUClientSessionCache(0),
 	}
 	b.dialContext = b.createDialContext(resolverAddresses)
 	b.resolvedConfig = b.createTLSConfig(host)
@@ -77,9 +94,9 @@ func newBootstrapperResolved(upsURL *url.URL, options *Options) (*bootstrapper,
 	return b, nil
 }
 
-// newBootstrapper initializes a new bootstrapper instance
-// address -- original resolver address string (i.e. tls://one.one.one.one:853)
-// options -- Upstream customization options
+// newBootstrapper initializes a new bootstrapper instance. u is the original
+// resolver address string (i.e. tls://one.one.one.one:853), options is the
+// upstream configuration options.
 func newBootstrapper(u *url.URL, options *Options) (b *bootstrapper, err error) {
 	resolvers := []*Resolver{}
 	if len(options.Bootstrap) != 0 {
@@ -103,18 +120,31 @@ func newBootstrapper(u *url.URL, options *Options) (b *bootstrapper, err error)
 		URL:       u,
 		resolvers: resolvers,
 		options:   options,
+		// Use the default capacity for the LRU cache.  It may be useful to
+		// store several caches since the user may be routed to different
+		// servers in case there's load balancing on the server-side.
+		sessionsCache: tls.NewLRUClientSessionCache(0),
 	}, nil
 }
 
-// dialHandler specifies the dial function for creating unencrypted TCP connections.
+// dialHandler describes the dial function for creating unencrypted network
+// connections to the upstream server.  Internally, this function will use the
+// supplied bootstrap DNS servers to resolve the upstream's IP address and only
+// then it will actually establish a connection.
 type dialHandler func(ctx context.Context, network, addr string) (net.Conn, error)
 
-// will get usable IP address from Address field, and caches the result
+// get is the main function of bootstrapper that does two crucial things.
+// First, it creates an instance of a dialHandler function that should be used
+// by the Upstream to establish a connection to the upstream DNS server.  This
+// dialHandler in a lazy manner resolves the DNS server IP address using the
+// bootstrap DNS servers supplied to this bootstrapper instance.  It will also
+// create an instance of *tls.Config that should be used for establishing an
+// encrypted connection for DoH/DoT/DoQ.
 func (n *bootstrapper) get() (*tls.Config, dialHandler, error) {
-	n.RLock()
+	n.guard.RLock()
 	if n.dialContext != nil && n.resolvedConfig != nil { // fast path
 		tlsConfig, dialContext := n.resolvedConfig, n.dialContext
-		n.RUnlock()
+		n.guard.RUnlock()
 		return tlsConfig.Clone(), dialContext, nil
 	}
 
@@ -123,22 +153,24 @@ func (n *bootstrapper) get() (*tls.Config, dialHandler, error) {
 	//
 
 	// get a host without port
-	addr := n.URL
-	host, port, err := net.SplitHostPort(addr.Host)
+	u := n.URL
+	host, port, err := net.SplitHostPort(u.Host)
 	if err != nil {
-		n.RUnlock()
-		return nil, nil, fmt.Errorf("bootstrapper requires port in address %s", addr.String())
+		n.guard.RUnlock()
+		return nil, nil, fmt.Errorf("bootstrapper requires port in address %s", u)
 	}
 
-	// if n.address's host is an IP, just use it right away
+	// if n.address's host is an IP, just use it right away.
 	ip := net.ParseIP(host)
 	if ip != nil {
-		n.RUnlock()
+		n.guard.RUnlock()
 
-		// Upgrade lock to protect n.resolved
 		resolverAddress := net.JoinHostPort(host, port)
-		n.Lock()
-		defer n.Unlock()
+
+		// Upgrade lock to protect n.resolvedConfig.
+		// TODO(ameshkov): rework, that's not how it should be done.
+		n.guard.Lock()
+		defer n.guard.Unlock()
 
 		n.dialContext = n.createDialContext([]string{resolverAddress})
 		n.resolvedConfig = n.createTLSConfig(host)
@@ -148,7 +180,7 @@ func (n *bootstrapper) get() (*tls.Config, dialHandler, error) {
 	// Don't lock anymore (we can launch multiple lookup requests at a time)
 	// Otherwise, it might mess with the timeout specified for the Upstream
 	// See here: https://github.com/AdguardTeam/dnsproxy/issues/15
-	n.RUnlock()
+	n.guard.RUnlock()
 
 	//
 	// if it's a hostname
@@ -182,23 +214,26 @@ func (n *bootstrapper) get() (*tls.Config, dialHandler, error) {
 		return nil, nil, fmt.Errorf("couldn't find any suitable IP address for host %s", host)
 	}
 
-	n.Lock()
-	defer n.Unlock()
+	n.guard.Lock()
+	defer n.guard.Unlock()
 
 	n.dialContext = n.createDialContext(resolved)
 	n.resolvedConfig = n.createTLSConfig(host)
 	return n.resolvedConfig, n.dialContext, nil
 }
 
-// createTLSConfig creates a client TLS config
+// createTLSConfig creates a client TLS config that will be used to establish
+// an encrypted connection for DoH/DoT/DoQ.
 func (n *bootstrapper) createTLSConfig(host string) *tls.Config {
 	tlsConfig := &tls.Config{
 		ServerName:            host,
 		RootCAs:               RootCAs,
 		CipherSuites:          CipherSuites,
+		ClientSessionCache:    n.sessionsCache,
 		MinVersion:            tls.VersionTLS12,
 		InsecureSkipVerify:    n.options.InsecureSkipVerify,
 		VerifyPeerCertificate: n.options.VerifyServerCertificate,
+		VerifyConnection:      n.options.VerifyConnection,
 	}
 
 	// Depending on the URL scheme, we choose what ALPN will be advertised by
@@ -209,7 +244,17 @@ func (n *bootstrapper) createTLSConfig(host string) *tls.Config {
 		//
 		// See https://github.com/ameshkov/dnslookup/issues/19.
 	case "https":
-		tlsConfig.NextProtos = []string{http2.NextProtoTLS, "http/1.1"}
+		httpVersions := n.options.HTTPVersions
+		if httpVersions == nil {
+			httpVersions = DefaultHTTPVersions
+		}
+
+		var nextProtos []string
+		for _, v := range httpVersions {
+			nextProtos = append(nextProtos, string(v))
+		}
+
+		tlsConfig.NextProtos = nextProtos
 	case "quic":
 		tlsConfig.NextProtos = compatProtoDQ
 	}
@@ -217,7 +262,8 @@ func (n *bootstrapper) createTLSConfig(host string) *tls.Config {
 	return tlsConfig
 }
 
-// createDialContext returns dialContext function that tries to establish connection with all given addresses one by one
+// createDialContext returns a dialHandler function that tries to establish the
+// connection to each of the provided addresses one by one.
 func (n *bootstrapper) createDialContext(addresses []string) (dialContext dialHandler) {
 	dialer := &net.Dialer{
 		Timeout: n.options.Timeout,
@@ -238,16 +284,38 @@ func (n *bootstrapper) createDialContext(addresses []string) (dialContext dialHa
 			conn, err := dialer.DialContext(ctx, network, resolverAddress)
 			elapsed := time.Since(start)
 			if err == nil {
-				log.Tracef("dialer has successfully initialized connection to %s in %s", resolverAddress, elapsed)
+				log.Tracef(
+					"dialer has successfully initialized connection to %s in %s",
+					resolverAddress,
+					elapsed,
+				)
 
 				return conn, nil
 			}
 
 			errs = append(errs, err)
 
-			log.Tracef("dialer failed to initialize connection to %s, in %s, cause: %s", resolverAddress, elapsed, err)
+			log.Tracef(
+				"dialer failed to initialize connection to %s, in %s, cause: %s",
+				resolverAddress,
+				elapsed,
+				err,
+			)
 		}
 
 		return nil, errors.List("all dialers failed", errs...)
 	}
 }
+
+// newContext creates a new context with deadline if needed.  If no timeout is
+// set cancel would be a simple noop.
+func (n *bootstrapper) newContext() (ctx context.Context, cancel context.CancelFunc) {
+	ctx = context.Background()
+	cancel = func() {}
+
+	if n.options.Timeout > 0 {
+		ctx, cancel = context.WithDeadline(ctx, time.Now().Add(n.options.Timeout))
+	}
+
+	return ctx, cancel
+}
